#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <signal.h>
#include <assert.h>
#include <fcntl.h>
#include <sys/time.h>
#include <uuid/uuid.h>

#include "msr-index.h"
#include "xenner.h"
#include "mm.h"

/* ------------------------------------------------------------------ */

static struct xenvm xen = {
    .pg_total = MB_TO_PG(64),
    .debug    = 1,
    .corefd   = -1,
    .name     = "xenner guest",
};

static int shutdown;
static int termsig;
static int timer;
static int pending_io;

/* ------------------------------------------------------------------ */

static void sighandle(int sig)
{
    switch (sig) {
    case SIGALRM:
	timer++;
	break;
    case SIGIO:
	pending_io++;
	break;
    case SIGINT:
	if (!shutdown) {
	    shutdown++;
	    break;
	}
	/* else fall through */
    default:
	fprintf(stderr, "%s: %s\n", __FUNCTION__, strsignal(sig));
	termsig = sig;
    }
}

static void sigsetup(void)
{
    struct sigaction act, old;

    /* install handlers */
    memset(&act,0,sizeof(act));
    act.sa_handler = sighandle;
    sigemptyset(&act.sa_mask);
    sigaction(SIGINT, &act, &old);
    sigaction(SIGALRM, &act, &old);
    sigaction(SIGIO, &act, &old);
}

void alive(void)
{
    static const char alive[] = ".oOo";
    static const int chars = sizeof(alive)/sizeof(alive[0]) -1;
    static int count;
    
    fprintf(stderr, "[%c]\r", alive[count++ % chars]);
}

uint64_t get_systime(void)
{
    uint64_t systime;
    struct timeval tv;

    gettimeofday(&tv, NULL);
    systime  = tv.tv_sec  * (uint64_t)1000000000;
    systime += tv.tv_usec * (uint64_t)1000;
    return systime;
}

/* ------------------------------------------------------------------ */

static void systime_to_tv(uint64_t systime, struct timeval *tv)
{
    tv->tv_sec  = systime / 1000000000;
    tv->tv_usec = (systime / 1000) % 1000000;
}

void setup_timer(struct xenvm *xen)
{
    struct itimerval it;
    int64_t next_tick;

    memset(&it, 0, sizeof(it));
    if (xen->period) {
	systime_to_tv(xen->period, &it.it_interval);
	systime_to_tv(xen->period, &it.it_value);
    }
    if (xen->oneshot) {
	uint64_t now = get_systime();
	next_tick = xen->oneshot - (now - xen->boot);
	if (next_tick <= 0) {
	    /* FIXME */
#if 0
	    d1printf("%s: now.o %.3f, now %.3f, one %.3f, next %.3f\n",
		     __FUNCTION__,
		     (xen->now - xen->boot) / 1000000000.0,
		     (now      - xen->boot) / 1000000000.0,
		     xen->oneshot           / 1000000000.0,
		     next_tick              / 1000000000.0);
#else
	    d2printf("%s: next_tick negative (%.6f sec), fixup\n",
		     __FUNCTION__, next_tick / 1000000000.0);
#endif
	    next_tick = xen->period;
	    if (0 == next_tick)
		next_tick = XEN_DEFAULT_PERIOD;
	}
	systime_to_tv(next_tick, &it.it_value);
    }

    d3printf("%s: itimer: val %ld.06%ld int %ld.06%ld\n", __FUNCTION__,
	     it.it_value.tv_sec,
	     it.it_value.tv_usec,
	     it.it_interval.tv_sec,
	     it.it_interval.tv_usec);
    setitimer(ITIMER_REAL, &it, NULL);
}

void update_time(struct xenvm *xen, int vcpu)
{
    uint64_t now, systime;

    now = get_systime();
    if (now - xen->now > 1000000)
	xen->update_time = 1;
    if (!xen->update_time)
	return;

    xen->update_time = 0;
    xen->now = now;
    systime = xen->now - xen->boot;

    d2printf("systime %" PRId64 ".%09" PRId64 "\n",
	     systime / 1000000000, systime % 1000000000);
    
    xen->xen_wc->wc_version++;
    xen->xen_wc->wc_sec  = now / 1000000000;
    xen->xen_wc->wc_nsec = now % 1000000000;
    xen->xen_wc->wc_version++;

    if (xen->vcpu[vcpu].time) {
	xen->vcpu[vcpu].time->version++;
	xen->vcpu[vcpu].time->system_time       = systime;
	xen->vcpu[vcpu].time->tsc_timestamp     = get_msr(xen, vcpu, MSR_IA32_TSC);
	xen->vcpu[vcpu].time->tsc_to_system_mul = xen->tsc_mul_frac;
	xen->vcpu[vcpu].time->tsc_shift         = xen->tsc_shift;
	xen->vcpu[vcpu].time->version++;
    }
}

int raise_event(struct xenvm *xen, int vcpu, int port)
{
    uint32_t *p32, *m32, *s32;
    uint64_t *p64, *m64, *s64;

    switch (xen->mode) {
    case XENMODE_32:
    case XENMODE_PAE:
	p32 = xen->shinfo.sh32->evtchn_pending;
	m32 = xen->shinfo.sh32->evtchn_mask;
	s32 = &xen->shinfo.sh32->vcpu_info[vcpu].evtchn_pending_sel;

	/* flag event port as pending */
	if (TESTBIT(p32, port))
	    break;
	SETBIT(p32, port);

	/* stop if event port is masked */
	if (TESTBIT(m32, port))
	    break;

	/* flag vcpu has events */
	if (TESTBIT(s32, port / 32))
	    break;
	SETBIT(s32, port / 32);
	*(xen->vcpu[vcpu].evtchn_upcall_pending) = 1;
	break;
    case XENMODE_64:
	p64 = xen->shinfo.sh64->evtchn_pending;
	m64 = xen->shinfo.sh64->evtchn_mask;
	s64 = &xen->shinfo.sh64->vcpu_info[vcpu].evtchn_pending_sel;

	/* flag event port as pending */
	if (TESTBIT(p64, port))
	    break;
	SETBIT(p64, port);

	/* stop if event port is masked */
	if (TESTBIT(m64, port))
	    break;

	/* flag vcpu has events */
	if (TESTBIT(s64, port / 64))
	    break;
	SETBIT(s64, port / 64);
	*(xen->vcpu[vcpu].evtchn_upcall_pending) = 1;
	break;
    }

#if 0
    fprintf(stderr, "%s: p 0x%02x, m 0x%02x, s 0x%02x, up %d um %d  |  %d\n",
	    __FUNCTION__, *p32, *m32, *s32,
	    *(xen->vcpu[vcpu].evtchn_upcall_pending),
	    *(xen->vcpu[vcpu].evtchn_upcall_mask),
	    port);
#endif
    return 0;
}

#define have_event(_x, _v) (*((_x)->vcpu[_v].evtchn_upcall_pending))
#define mask_event(_x, _v) (*((_x)->vcpu[_v].evtchn_upcall_mask))

static int send_event(struct xenvm *xen, int vcpu)
{
    return have_event(xen, vcpu) && !mask_event(xen, vcpu);
}

void wait_event(struct xenvm *xen, int vcpu)
{
    sigset_t suspend;
    sigset_t block;
    sigset_t orig;
    
    *(xen->vcpu[0].evtchn_upcall_mask) = 0;

    if (have_event(xen, vcpu))
	return;
    
    sigfillset(&suspend);
    sigemptyset(&block);
    sigdelset(&suspend, SIGALRM);
    sigaddset(&block,   SIGALRM);
    sigdelset(&suspend, SIGIO);
    sigaddset(&block,   SIGIO);

    sigdelset(&suspend, SIGINT);

    sigprocmask(SIG_BLOCK, &block, &orig);
    if (!timer && !pending_io) {
	sigsuspend(&suspend);
	xen->update_time = 1;
    }
    sigprocmask(SIG_SETMASK, &orig, NULL);
}

int xenner_register_cb(struct xenvm *xen, enum xenner_cb_type type,
		       char *name, int handle, void *cb_data,
		       int (*callback)(struct xenvm *xen, int fd, void *data))
{
    struct xenner_callback *cb;

    cb = malloc(sizeof(*cb));
    memset(cb,0,sizeof(*cb));
    cb->handle   = handle;
    cb->name     = name;
    cb->cb_data  = cb_data;
    cb->callback = callback;
    list_add_tail(&cb->list, &xen->callbacks[type]);

    switch (type) {
    case XENNER_CB_IOFD:
	/* setup stuff so we get SIGIO on data */
	if (0 != fcntl(cb->handle, F_SETFL, O_ASYNC | O_NONBLOCK)) {
	    perror("fcntl F_SETFL");
	    return -1;
	}
	if (0 != fcntl(cb->handle, F_SETOWN, getpid())) {
	    perror("fcntl F_SETOWN");
	    return -1;
	}
	if (0 != fcntl(cb->handle, F_SETSIG, SIGIO)) {
	    perror("fcntl F_SETSIG");
	    return -1;
	}
	break;
    default:
	/* nothing special to do */
	break;
    }
    return 0;
}

static int callback_iofds(struct xenvm *xen)
{
    struct list_head *item;
    struct xenner_callback *cb;
    int count;

    list_for_each(item, &xen->callbacks[XENNER_CB_IOFD]) {
	cb = list_entry(item, struct xenner_callback, list);
	count = cb->callback(xen, cb->handle, cb->cb_data);
    }
    return 0;
}

int xenner_callback_evtchn(struct xenvm *xen, int port)
{
    struct list_head *item;
    struct xenner_callback *cb;

    list_for_each(item, &xen->callbacks[XENNER_CB_EVTCHN]) {
	cb = list_entry(item, struct xenner_callback, list);
	if (cb->handle != port)
	    continue;
	d2printf("%s: calling %s\n", __FUNCTION__, cb->name);
	cb->callback(xen, cb->handle, cb->cb_data);
	return 0;
    }
    return -1;
}

/* ------------------------------------------------------------------ */

void need_regs(struct xenvm *xen)
{
    if (xen->regs_up2date)
	return;
    kvm_get_regs(xen->kvm, 0, &xen->regs);
    xen->regs_up2date = 1;
}

void need_sregs(struct xenvm *xen)
{
    if (xen->sregs_up2date)
	return;
    kvm_get_sregs(xen->kvm, 0, &xen->sregs);
    xen->sregs_up2date = 1;
}

void flush_regs(struct xenvm *xen)
{
    assert(xen->regs_up2date);
    xen->regs_dirty = 1;
}

void flush_sregs(struct xenvm *xen)
{
    assert(xen->sregs_up2date);
    xen->sregs_dirty = 1;
}

void xenner_cleanup(struct xenvm *xen)
{
    fprintf(stderr, "--- statistics ---\n");
    hypercall_stats(xen);

    fprintf(stderr, "--- cleaning up ---\n");
    vmcore_fini(xen);
    console_fini(xen);
    xenstore_fini(xen);
    evtchn_fini(xen);
}

/* ------------------------------------------------------------------ */

static void print_stack_32(struct xenvm *xen, char *name, uint32_t esp, uint32_t *ptr)
{
    int i;
    
    fprintf(stderr, "--- %s stack trace ---\n", name);
    for (i = 0; i < 128; i++) {
	if (0 == addr_offset(esp + 4*i))
	    break;
	if (0 == (i % 4))
	    fprintf(stderr, "%08" PRIx32 ":", esp + 4*i);
	fprintf(stderr, " %08" PRIx32, ptr[i]);
	if (3 == (i % 4))
	    fprintf(stderr, "\n");
    }
    if (0 != (i % 4))
	fprintf(stderr, "\n");
}

static void print_stack_64(struct xenvm *xen, char *name, uint64_t esp, uint64_t *ptr)
{
    int i;
    
    fprintf(stderr, "--- %s stack trace ---\n", name);
    for (i = 0; i < 128; i++) {
	if (0 == addr_offset(esp + 8*i))
	    break;
	if (0 == (i % 4))
	    fprintf(stderr, "%016" PRIx64 ":", esp + 8*i);
	fprintf(stderr, " %016" PRIx64, ptr[i]);
	if (3 == (i % 4))
	    fprintf(stderr, "\n");
    }
    if (0 != (i % 4))
	fprintf(stderr, "\n");
}

void print_guest_stack(struct xenvm *xen, uint64_t rsp)
{
    if (xen->mode == XENMODE_64) {
	uint64_t *data = guest_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_64(xen, "guest64", rsp, data);
    } else {
	uint32_t *data = guest_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_32(xen, "guest32", rsp, data);
    }
}

void print_emu_stack(struct xenvm *xen, uint64_t rsp)
{
    if (xen->mode == XENMODE_64) {
	uint64_t *data = emu_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_64(xen, "emu64", rsp, data);
    } else {
	uint32_t *data = emu_vaddr_to_ptr(xen, rsp);
	if (data)
	    print_stack_32(xen, "emu32", rsp, data);
    }
}

void vm_kill(struct xenvm *xen, char *reason, uint64_t guest_rsp)
{
    banner_print(xen, reason);
    fprintf(stderr, "--- registers ---\n");
    kvm_show_regs(xen->kvm, 0);
    print_msrs(xen, 0);

    need_regs(xen);
    if (xen->regs.rsp > xen->emu_va && xen->regs.rsp < xen->emu_stack)
	print_emu_stack(xen, xen->regs.rsp);
    else if (!guest_rsp && xen->regs.rsp)
	guest_rsp = xen->regs.rsp;

    if (guest_rsp)
	print_guest_stack(xen, guest_rsp);	

    xenner_cleanup(xen);
    exit(1);
}

/* ------------------------------------------------------------------ */
/* functions                                                          */

static int cb_inb(void *opaque, uint16_t addr, uint8_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xff;
    return 0;
}

static int cb_inw(void *opaque, uint16_t addr, uint16_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xffff;
    return 0;
}

static int cb_inl(void *opaque, uint16_t addr, uint32_t *value)
{
    struct xenvm *xen = opaque;

    d2printf("io: %s, addr 0x%" PRIx16 "\n",
	     __FUNCTION__, addr);
    *value = 0xffffffff;
    return 0;
}

static int cb_outb(void *opaque, uint16_t addr, uint8_t value)
{
    struct xenvm *xen = opaque;
    char *msg;

    switch (addr) {
#if 1 /* handled by emu now, shouldn't happen any more */
    case 0xe1: /* illegal instruction */
	vm_kill(xen, "illegal instruction (obsolete)", 0);
	// do_instr_emu(xen);
	break;
    case 0xe2: /* page fault */
	vm_kill(xen, "page fault (obsolete)", 0);
	break;
    case 0xe3: /* general protection */
	vm_kill(xen, "general protection (obsolete)", 0);
//	do_general_protection(xen);
	break;
    case 0xe5: /* event callback (irq 0x20) */
	vm_kill(xen, "event callback (obsolete)", 0);
	break;
#endif

    case 0xe0: /* hypercall */
	do_xen_hypercall(xen, value);
	break;
    case 0xe4: /* division by zero */
	vm_kill(xen, "division by zero", 0);
	break;

    case 0xea: /* emu: print string */
	need_regs(xen);
	msg = emu_vaddr_to_ptr(xen, xen->regs.rdx);
	conemu_print(xen, msg);
	break;
    case 0xeb: /* emu: panic */
	vm_kill(xen, "emu: panic()", 0);
	break;

    case 0xec: /* debug trap */
	do_debug_trap(xen);
	break;

    case 0xef: /* unhandled interrupt */
	vm_kill(xen, "unknown irq", 0);
	break;
    default:
	d2printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx8 "\n",
		__FUNCTION__, addr, value);
	break;
    }
    return 0;
}

static int cb_outw(void *opaque, uint16_t addr, uint16_t value)
{
    struct xenvm *xen = opaque;

    d1printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx16 "\n",
	     __FUNCTION__, addr, value);
    return 0;
}

static int cb_outl(void *opaque, uint16_t addr, uint32_t value)
{
    struct xenvm *xen = opaque;

    d1printf("io: %s, addr 0x%" PRIx16 " value 0x%" PRIx32 "\n",
	     __FUNCTION__, addr, value);
    return 0;
}

static int cb_debug(void *opaque, int vcpu)
{
    struct xenvm *xen = opaque;
    
    d1printf("callback: %s\n", __FUNCTION__);
    return 0;
}

static int cb_halt(void *opaque, int vcpu)
{
    struct xenvm *xen = opaque;

    d1printf("callback: %s\n", __FUNCTION__);
    vm_kill(xen, "DEBUG: halt", 0);
    sleep(1);
    return 0;
}

static int cb_shutdown(void *opaque, int vcpu)
{
    struct xenvm *xen = opaque;

    d1printf("callback: %s\n", __FUNCTION__);
    vm_kill(xen, "crashed", 0);
    return 0;
}

static int cb_io_window(void *opaque)
{
    struct xenvm *xen = opaque;

    d2printf("callback: %s\n", __FUNCTION__);
    return 0;
}

static int cb_try_push_interrupts(void *opaque)
{
    struct xenvm *xen = opaque;
    
    d3printf("callback: %s\n", __FUNCTION__);

    if (shutdown) {
	static int once = 1;
	if (once)
	    xenstore_shutdown(xen, "halt");
	once = 0;
    }

    if (pending_io) {
	pending_io = 0;
	callback_iofds(xen);
    }

    if (timer && xen->virq[VIRQ_TIMER]) {
	timer = 0;
	raise_event(xen, 0, xen->virq[VIRQ_TIMER]);
    }

    if (send_event(xen, 0)) {
//	fprintf(stderr, "%s: irq\n", __FUNCTION__);
	kvm_inject_irq(xen->kvm, 0, 0x20);
    }

    return 0;
}

static void cb_post_kvm_run(void *opaque, int vcpu)
{
    struct xenvm *xen = opaque;

    d3printf("callback: %s\n", __FUNCTION__);
}

static int cb_pre_kvm_run(void *opaque, int vcpu)
{
    struct xenvm *xen = opaque;

    d3printf("callback: %s\n", __FUNCTION__);
    if (termsig)
	vm_kill(xen, "termsig", 0);

    /* timekeeping */
    update_time(xen, vcpu);

    /* registers */
    if (xen->regs_dirty) {
	kvm_set_regs(xen->kvm, 0, &xen->regs);
	xen->regs_dirty = 0;
    }
    if (xen->sregs_dirty) {
	kvm_set_sregs(xen->kvm, 0, &xen->sregs);
	xen->sregs_dirty = 0;
    }
    xen->regs_up2date = 0;
    xen->sregs_up2date = 0;
    return 0;
}

/* ------------------------------------------------------------------ */
/* main                                                               */

static struct kvm_callbacks xenner_cb = {
    .inb                 = cb_inb,
    .inw                 = cb_inw,
    .inl                 = cb_inl,
    .outb                = cb_outb,
    .outw                = cb_outw,
    .outl                = cb_outl,
    .debug               = cb_debug,
    .halt                = cb_halt,
    .shutdown            = cb_shutdown,
    .io_window           = cb_io_window,
    .try_push_interrupts = cb_try_push_interrupts,
    .post_kvm_run        = cb_post_kvm_run,
    .pre_kvm_run         = cb_pre_kvm_run,
};

static void usage(FILE *fp)
{
    fprintf(fp,
	    "\n"
	    "xenner -- run xen kernels in kvm\n"
	    "\n"
	    "usage: xenner [options]\n"
	    "options:\n"
	    "   -h            print this text\n"
	    "   -q            be quiet (debuglevel 0)\n"
	    "   -d            increase debuglevel\n"
	    "   -k <kernel>   xenified kernel image\n"
	    "   -r <ramdisk>  ramdisk\n"
	    "   -a <append>   kernel command line\n"
	    "   -m <memory>   vm memory in megabytes\n"
	    "   -c <name>     pick color scheme (off, light, dark)\n"
	    "   -l <file>     log messages to file (and stderr)\n"
	    "   -L <file>     log messages only to file\n"
	    "\n"
	    "   -i <nr>       domain id\n"
	    "   -n <string>   domain name\n"
	    "   -u <uuid>     domain uuid\n"
	    "\n"
	    "   -D <file>     disk image\n"
	    "   -C <file>     cdrom image\n"
	    "   -F <type>     framebuffer type (vnc, sdl, qemu)\n"
	    "   -N <mac>,<br> network bridge and mac address\n"
	    "\n"
	    "-- \n"
	    "(c) 2007 Gerd Hoffmann <kraxel@redhat.com>\n"
	    "\n");
}

int main(int argc, char *argv[])
{
    char *colors = NULL;
    char *uuid = NULL;
    int c;

    xen.domid = getpid();
    INIT_LIST_HEAD(&xen.callbacks[XENNER_CB_IOFD]);
    INIT_LIST_HEAD(&xen.callbacks[XENNER_CB_EVTCHN]);

    for (;;) {
        if (-1 == (c = getopt(argc, argv,
			      "hdqk:K:r:a:m:c:L:l:"
			      "i:n:u:"
			      "D:C:F:N:")))
            break;
        switch (c) {
	case 'q':
	    xen.debug = 0;
	    break;
	case 'd':
	    xen.debug++;
	    break;
	case 'k':
	    xen.kernel = optarg;
	    break;
	case 'K':
	    parse_asm(optarg);
	    break;
	case 'r':
	    xen.ramdisk = optarg;
	    break;
	case 'a':
	    xen.cmdline = optarg;
	    break;
	case 'm':
	    xen.pg_total = MB_TO_PG(atoi(optarg));
	    break;
	case 'c':
	    colors = optarg;
	    break;
	case 'L':
	    xen.nostderr = 1;
	    /* fall through */
	case 'l':
	    xen.logfile = fopen(optarg, "w");
	    if (NULL == xen.logfile) {
		fprintf(stderr, "can't open %s: %s\n", optarg, strerror(errno));
		exit(1);
	    }
	    break;

	case 'i':
	    xen.domid = atoi(optarg);
	    break;
	case 'n':
	    snprintf(xen.name, sizeof(xen.name), "%s", optarg);
	    break;
	case 'u':
	    uuid = optarg;
	    break;

	case 'D':
	    xen.disk = optarg;
	    break;
	case 'C':
	    xen.cdrom = optarg;
	    break;
	case 'F':
	    xen.vfb = optarg;
	    break;
	case 'N':
	    xen.network = optarg;
	    break;

        case 'h':
            usage(stdout);
            exit(0);
        default:
            usage(stderr);
            exit(1);
        }
    }
    loginit(colors);

    if (NULL == uuid) {
	uuid_t u;
	strcpy(xen.vm, "/vm/");
	uuid_generate(u);
	uuid_unparse(u, xen.vm+4);
    } else {
	sprintf(xen.vm, "/vm/%.36s", uuid);
    }

    /* init kvm */
    xen.kvm = kvm_init(&xenner_cb, &xen);
    if (!xen.kvm) {
	fprintf(stderr, "kvm_init failed\n");
	goto err;
    }
    kvm_disable_irqchip_creation(xen.kvm);
    
    /* create vm */
    xen.pg_total &= ~((1<<10)-1);  /* round to 4MB border */
    if (vmcore_init(&xen) < 0)
	goto err;
    if (kvm_create_vm(xen.kvm) < 0) {
	fprintf(stderr, "kvm_create_vm failed\n");
	goto err;
    }

#ifndef KVM_CAP_SET_TSS_ADDR
#error linux/kvm.h is too old (need 49+).
#endif
    if (kvm_set_tss_addr(xen.kvm, xen.pg_total * PAGE_SIZE) < 0) {
	fprintf(stderr, "kvm_set_tss_addr failed (%s)\n", strerror(errno));
	fprintf(stderr, "Old kvm version maybe?  I need 49+.\n");
	goto err;
    }

    if (kvm_register_userspace_phys_mem(xen.kvm, 0, xen.memory,
					xen.pg_total * PAGE_SIZE, 1) < 0) {
	fprintf(stderr, "kvm_register_userspace_phys_mem failed\n");
	goto err;
    }
    if (kvm_create_vcpu(xen.kvm, 0) < 0) {
	fprintf(stderr, "kvm_create_vcpu failed\n");
	goto err;
    }

    need_regs(&xen);
    need_sregs(&xen);

    banner_print(&xen, "setup");
    sigsetup();
    if (0 != evtchn_init(&xen))
	goto err;
    if (0 != xenstore_init_early(&xen))
	goto err;
    console_init(&xen);

    if (0 != domain_builder(&xen)) {
	fprintf(stderr, "domain builder failed\n");
	goto err;
    }
    flush_regs(&xen);
    flush_sregs(&xen);

    xenstore_init_late(&xen);

#if 0
    kvm_set_regs(xen.kvm, 0, &xen.regs);
    kvm_set_sregs(xen.kvm, 0, &xen.sregs);
    kvm_show_regs(xen.kvm, 0);
#endif

    setup_cpuid(&xen);

    tsc_calibrate(&xen.tsc_shift, &xen.tsc_mul_frac);
    xen.boot = get_systime();
    xen.period = XEN_DEFAULT_PERIOD;
    setup_timer(&xen);

    vmcore_write_headers(&xen);

    banner_print(&xen, "boot");
    kvm_run(xen.kvm, 0);
    return 0;

err:
    if (xen.kvm)
	kvm_finalize(xen.kvm);
    if (-1 != xen.corefd)
	unlink(xen.corefile);
    return 1;
}
